package com.bergerkiller.bukkit.nolagg.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import net.minecraft.server.ChunkProviderServer;
import net.minecraft.server.WorldServer;

import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.entity.Entity;
import org.bukkit.entity.ExperienceOrb;
import org.bukkit.entity.Item;

import com.all.bukkit.threadcheck.ThreadCheck;
import com.bergerkiller.bukkit.nolagg.ItemHandler;
import com.bergerkiller.bukkit.nolagg.SpawnHandler;
import com.bergerkiller.bukkit.nolagg.StackFormer;
import com.bergerkiller.bukkit.nolagg.Util;

public class WorldMetaData {
	public static double entityListGenDuration = 0;
	public static double entitySpawnLimitDuration = 0;
	public static double itemStackDuration = 0;
	public static double chunkUpdateDuration = 0;
	public static double chunkUnloadDuration = 0;
	private static final double useOld = 0.5;
	public static double getTotalUpdateTime() {
		return entityListGenDuration + entitySpawnLimitDuration + itemStackDuration + chunkUpdateDuration + chunkUnloadDuration;
	}
	
	private static final List<WorldMetaData> worlds = new ArrayList<WorldMetaData>();
	private static final HashMap<UUID, WorldMetaData> worlddata = new HashMap<UUID, WorldMetaData>();
	public static void init() {
		for (WorldServer world : Util.getWorlds()) get(world);
	}
	public static void deinit() {
		synchronized (worlddata) {
			for (WorldMetaData dat : worlddata.values()) {
				dat.remove();
			}
			worlddata.clear();
			worlds.clear();
		}
	}
	public static void remove(final World world) {
		synchronized (worlddata) {
			WorldMetaData dat = worlddata.remove(world.getUID());
			if (dat != null) {
				worlds.remove(dat);
				dat.remove();
			}
		}
	}
	public static WorldMetaData get(final World world) {
		if (world == null) return null;
		synchronized (worlddata) {
			WorldMetaData dat = worlddata.get(world.getUID());
			if (dat == null) {
				dat = new WorldMetaData(world);
				worlddata.put(world.getUID(), dat);
				worlds.add(dat);
			}
			return dat;
		}
	}
	public static WorldMetaData get(final net.minecraft.server.World world) {
		return get(world.getWorld());
	}
	public static void updateAll() {
		if (!ThreadCheck.check(null)) return;
		SpawnHandler.reset();
		synchronized (worlddata) {
			long time = System.currentTimeMillis();
			long newtime;
			for (WorldMetaData dat : worlds) dat.updateEntities();
			newtime = System.currentTimeMillis();		
			entityListGenDuration = Util.useOld(entityListGenDuration, newtime - time, useOld);
			time = newtime;
			for (WorldMetaData dat : worlds) dat.updateSpawnLimits();
			newtime = System.currentTimeMillis();
			entitySpawnLimitDuration = Util.useOld(entitySpawnLimitDuration, newtime - time, useOld);
			time = newtime;		
			for (WorldMetaData dat : worlds) dat.updateStackForming();		
			newtime = System.currentTimeMillis();
			itemStackDuration = Util.useOld(itemStackDuration, newtime - time, useOld);
			time = newtime;	
			for (WorldMetaData dat : worlds) dat.updateChunks();
			newtime = System.currentTimeMillis();
			chunkUpdateDuration = Util.useOld(chunkUpdateDuration, newtime - time, useOld);
			time = newtime;	
			unloadChunks();
			newtime = System.currentTimeMillis();
			chunkUnloadDuration = Util.useOld(chunkUnloadDuration, newtime - time, useOld);
		}
	}
	private static final ArrayList<ChunkMetaData> toUnload = new ArrayList<ChunkMetaData>();
	private void updateEntities() {
		//update entity list
		this.entitylist.clear();
		this.itemlist.clear();
		this.orblist.clear();
		Util.fillEntities(this.getNative(), this.entitylist, this.itemlist, this.orblist);
	}
	private void updateSpawnLimits() {
		SpawnHandler.update(this.getNative(), this.entitylist);
	}
	private void updateStackForming() {
		StackFormer.update(this.itemlist, this.orblist);
	}
	private void updateChunks() {
		//update chunks and unload them if needed
		synchronized (this.chunkdata) {
			if (!this.chunkdata.isEmpty()) {
				final long time = System.currentTimeMillis() - ChunkMetaData.unloadDelay;
				Iterator<ChunkMetaData> it = this.chunkdata.values().iterator();
				ChunkMetaData dat;
				while (it.hasNext()) {
					dat = it.next();
					if (!dat.isLoaded()) {
						it.remove();
					} else if (!dat.isUnloadable() || dat.isPlayerNear()) {
						dat.update();
						dat.touch();
					} else if (dat.isExpired(time)) {
						toUnload.add(dat);
					} else {
						dat.update();
					}
				}

			}
		}
	}
	private static void unloadChunks() {
		for (ChunkMetaData data : toUnload) {
			if (data.unload()) {
				data.getMetaWorld().removeChunk(data);
			} else {
				data.update();
			}
		}
		toUnload.clear();
	}
	
	
	
	public static int getTotalChunkCount() {
		int c = 0;
		synchronized (worlddata) {
			for (WorldMetaData wmd : worlddata.values()) c += wmd.getChunkCount();	
		}
		return c;
	}
	public static int getTotalUnloadableChunkCount() {
		int c = 0;
		synchronized (worlddata) {
			for (WorldMetaData wmd : worlddata.values()) {
				synchronized (wmd.chunkdata) {
					for (ChunkMetaData cmd : wmd.chunkdata.values()) {
					    if (cmd.isUnloadable()) c++;
					}
				}
			}
		}
		return c;
	}
	public static int getTotalHiddenItemCount() {
		int c = 0;
		synchronized (worlddata) {
			for (WorldMetaData wmd : worlddata.values()) {
				synchronized (wmd.chunkdata) {
					for (ChunkMetaData cmd : wmd.chunkdata.values()) {
						c += cmd.getHiddenItemCount();
					}
				}
			}
		}
		return c;
	}

	public static void clearAllItems() {
		synchronized (worlddata) {
			for (WorldMetaData dat : worlddata.values()) dat.clearItems();
		}
	}
	public static boolean handleItemSpawn(Item item, boolean doIgnoreCheck) {
		return ChunkMetaData.get(item.getLocation()).handleItemSpawn(item, doIgnoreCheck);
	}
    public static boolean removeSpawnedItem(Item item) {
    	if (!ThreadCheck.check(null)) return false;
    	if (ItemHandler.maxItemsPerChunk <= 0) return false;
    	WorldMetaData dat = get(item.getWorld());
    	if (dat.getChunk(item.getLocation()).removeSpawnedItem(item)) {
    		return true;
    	} else {
    		//loop through chunks
    		synchronized (dat.chunkdata) {
        		for (ChunkMetaData cdat : dat.chunkdata.values()) {
        			if (cdat.removeSpawnedItem(item)) return true;
        		}
    		}
    		return false;
    	}
    }
	
	private WorldMetaData(final World world) {
		this.world = (CraftWorld) world;
		this.chunkdata = new HashMap<Long, ChunkMetaData>();
		this.entitylist = new ArrayList<Entity>();
		this.itemlist = new ArrayList<Item>();
		this.orblist = new ArrayList<ExperienceOrb>();
		this.add();
	}
		
	public final CraftWorld world;
	private final Map<Long, ChunkMetaData> chunkdata;
	private final List<Entity> entitylist;
	private final List<Item> itemlist;
	private final List<ExperienceOrb> orblist;
	
	public ChunkProviderServer getChunkProvider() {
		return this.getNative().chunkProviderServer;
	}
	public WorldServer getNative() {
		return this.world.getHandle();
	}
	
	public int getHiddenItemCount() {
		int c = 0;
		synchronized (this.chunkdata) {
			for (ChunkMetaData cmd : this.chunkdata.values()) {
				c += cmd.getHiddenItemCount();
			}
		}
		return c;
	}
	public void clearItems() {
		if (!ThreadCheck.check(null)) return;
		synchronized (this.chunkdata) {
			for (ChunkMetaData cmd : this.chunkdata.values()) {
				cmd.clearItems();
			}
		}
	}

	public List<Entity> getEntities() {
		return this.entitylist;
	}
	public List<Item> getItems() {
		return this.itemlist;
	}
	public List<ExperienceOrb> getOrbs() {
		return this.orblist;
	}
	
	public ChunkMetaData getChunk(final Location location) {
		return this.getChunk(location.getBlockX() >> 4, location.getBlockZ() >> 4);
	}
	public ChunkMetaData getChunk(final net.minecraft.server.Chunk chunk) {
		return this.getChunk(chunk.x, chunk.z);
	}
 	public ChunkMetaData getChunk(final Chunk chunk) {
		return this.getChunk(chunk.getX(), chunk.getZ());
	}
	public ChunkMetaData getChunk(final int x, final int z) {
		if (!ThreadCheck.check(null)) return null;
		final long key = Util.toLong(x, z);
		synchronized (this.chunkdata) {
			ChunkMetaData dat = this.chunkdata.get(key);
			if (dat == null) {
				dat = new ChunkMetaData(this, x, z);
				this.chunkdata.put(key, dat);
			}
			return dat;
		}
	}
	public int getChunkCount() {
		synchronized (this.chunkdata) {
			return this.chunkdata.size();
		}
	}
	
	protected void removeChunk(final ChunkMetaData dat) {
		synchronized (this.chunkdata) {
			this.chunkdata.remove(Util.toLong(dat.getX(), dat.getZ()));
		}
	}
	public boolean removeChunk(final int x, final int z) {
		synchronized (this.chunkdata) {
			ChunkMetaData dat = this.chunkdata.get(Util.toLong(x, z));
			if (dat != null) {
				dat.remove();
				return true;
			} else {
				return false;
			}
		}
	}
	

	private void add() {
		for (Chunk c : world.getLoadedChunks()) getChunk(c);
	}
	private void remove() {
		synchronized (this.chunkdata) {
			Iterator<ChunkMetaData> it = this.chunkdata.values().iterator();
			ChunkMetaData dat;
			while (it.hasNext()) {
				dat = it.next();
				dat.remove(false);
				it.remove();
			}
		}
		//other unload code here
	}

}
